#===================================================================#
# Python algorithm to perform T-Station to VELO relative alignment
#===================================================================#
# @author: Steven Blusk,  sblusk@phy.syr.edu
# @ date: 2007-03-02
#===================================================================#
#
# Note: You MUST have a local copy of LHCBCOND.db to run this code in the
# current working directory.
#

import gaudimodule
gaudimodule.loaddict('LHCbKernelDict')

from os import environ,system
import myConstants
import RanCons
import random
import sys
import getopt
import conddbui
from conddbui import cool
import math
from ROOT import *

from AlignmentTools import *

#-------------- Some convenient definitions -----------------
ROOT = gaudimodule.gbl.ROOT
std = gaudimodule.gbl.std
XYZPoint = ROOT.Math.XYZPoint
XYZVector = ROOT.Math.XYZVector
Measurement = gaudimodule.gbl.LHCb.Measurement
Track = gaudimodule.gbl.LHCb.Track
State = gaudimodule.gbl.LHCb.State
pi = math.pi
seedlocation = 'Rec/Track/Tsa'
velolocation = 'Rec/Track/PreparedVelo'
#------------------------------------------------------------


#def makeDB(dbase):
#    db_name = "LHCBCOND"
#    source_file = environ["LHCb_release_area"] + "/DBASE/Det/SQLDDDB/v1r0/db/LHCBCOND.db"
#    source = "sqlite://;schema=%s;dbname=%s"%(source_file, db_name)
#    dest_file = environ["PWD"] + '/' + dbase
#    dest = "sqlite://;schema=%s;dbname=%s"%(dest_file, db_name)
#    conddbui.CondDB.dropDatabase(dest)
#    copy(source, dest, tags = ['DC06'])

        
def SetupROOT():
    gStyle.SetOptFit(1)
    gStyle.SetOptStat(0)

    
class VeloTAlignment:

    def __init__(self, optsFile, name = 'VeloTAlignment'):
        self.name = name
        self.optsFile = optsFile

        appMgr = gaudimodule.AppMgr(outputlevel=3, joboptions=optsFile)

        appMgr.initialize()
        self.gaudi = appMgr

        self.evt = appMgr.evtsvc()
        self.det = appMgr.detsvc()
        self.sel = appMgr.evtsel()

#       self.extrapolatorTool = appMgr.toolsvc().create('TrackLinearExtrapolator', interface='ITrackExtrapolator')

        self.zLast = 9500.0
        self.zFirst = 0.0
        self.zMagnet = 5294.0


        self.numIter = 8

        # Book Histograms
        self.pkx =  TH1F('pkx', 'Px Kick', 200, -4.0, 4.0)
        self.dxh =  TH1F('DeltaX', 'Delta X', 200, -50.0, 50.0)
        self.dyh =  TH1F('DeltaY', 'Delta Y', 200, -50.0, 50.0)
        self.dysh =  TH1F('DeltaYSlope', 'Delta Y Slope', 200, -20, 20)
        self.dxsh =  TH1F('DeltaXSlope', 'Delta X Slope', 200, -20, 20)
        self.dxswh =  TH1F('dxsw', 'Delta X Slope', 100, -50, 50)
        self.dphih = TH1F('DeltaPhi', 'Delta Phi', 200, -200, 200)
        self.dzxh =  TH1F('DeltaZX', 'Delta Z (X view)', 200, -1000., 1000.)
        self.dzyh =  TH1F('DeltaZY', 'Delta Z (Y view)', 200, -1000., 1000.)
        self.dzh =   TH1F('DeltaZ', 'Delta Z', 200, -1000., 1000.)
        self.zinth =   TH1F('Zint', 'Z Intersection', 400, 4900., 5700.)
        self.momh = TH1F('Momentum', 'P (GeV)', 200, 0., 100.)
        self.type = TH1F('type', 'Measurement Type',10,0.0,10.0)
        self.posxy = TH2F('XYpos','Y vs X Position',100,-1000.,1000.,100,-1000.,1000.)
        self.posx = TH1F('Xpos','X Position at Z=8000 mm',300,-3000.,3000.)
        self.posy = TH1F('Ypos','Y Position at Z=8000 mm',300,-3000.,3000.)
        self.nIT = TH1F('nITHits','"F#IT Hits',20,0.,20.)
        self.nOT = TH1F('nOTHits','"F#OT Hits',40,0.,40.)
        self.nT = TH1F('nTHits','"F#IT Hits',60,0.,60.)

# Look at some quantities as a function of charge
        self.dxhp =  TH1F('DeltaXP', 'Delta X (+)', 200, -50.0, 50.0)
        self.dyhp =  TH1F('DeltaYP', 'Delta Y (+)', 200, -50.0, 50.0)
        self.dyshp =  TH1F('DeltaYSlopeP', 'Delta Y Slope (+)', 200, -20, 20)
        self.dxshp =  TH1F('DeltaXSlopeP', 'Delta X Slope (+)', 200, -20, 20)
        self.dphihp = TH1F('DeltaPhiP', 'Delta Phi (+)', 200, -200, 200)
        self.dzxhp =  TH1F('DeltaZXP', 'Delta Z (X view) (+)', 200, -1000., 1000.)
        self.dzyhp =  TH1F('DeltaZYP', 'Delta Z (Y view) (+)', 200, -1000., 1000.)
        self.dzhp =   TH1F('DeltaZP', 'Delta Z (+)', 200, -1000., 1000.)
        self.zinthp =   TH1F('ZintP', 'Z Intersection (+)', 400, 4900., 5700.)

        self.dxhn =  TH1F('DeltaXN', 'Delta X (-)', 200, -50.0, 50.0)
        self.dyhn =  TH1F('DeltaYN', 'Delta Y (-)', 200, -50.0, 50.0)
        self.dyshn =  TH1F('DeltaYSlopeN', 'Delta Y Slope (-)', 200, -20, 20)
        self.dxshn =  TH1F('DeltaXSlopeN', 'Delta X Slope (-)', 200, -20, 20)
        self.dphihn = TH1F('DeltaPhiN', 'Delta Phi (-)', 200, -200, 200)
        self.dzxhn =  TH1F('DeltaZXN', 'Delta Z (X view) (-)', 200, -1000., 1000.)
        self.dzyhn =  TH1F('DeltaZYN', 'Delta Z (Y view) (-)', 200, -1000., 1000.)
        self.dzhn =   TH1F('DeltaZN', 'Delta Z (-)', 200, -1000., 1000.)
        self.zinthn =   TH1F('ZintN', 'Z Intersection (-)', 400, 4900., 5700.)


        if(myConstants.myFitFunc == 'Gauss2'):
            self.fitFunc = TF1('myfunc', MyGauss2(),-20.0, 20.0, 7)
            self.MakeFitHistos2()
            print "Initializing Fitting Function ",myConstants.myFitFunc
        elif(myConstants.myFitFunc == 'Gauss3'):
            self.fitFunc = TF1('myfunc', MyGauss3(),-20.0, 20.0, 9)
            self.MakeFitHistos3()

        print "--------------------------------------------------------------"
        print "    Initializing Fitting Function ",myConstants.myFitFunc
        print "--------------------------------------------------------------"

        SetupROOT()
        
    def configure(self, seedloc, veloloc, numEv):
        self.seedlocation = seedloc
        self.velolocation = veloloc
        self.numEvProcess = numEv

        appMgr = self.gaudi

        self.updMgrSvc = appMgr.service('UpdateManagerSvc','IUpdateManagerSvc')
        self.magFieldSvc = appMgr.service('MagneticFieldSvc','IMagneticFieldSvc')
        self.condAccSvc = appMgr.service('LHCBCOND','ICondDBAccessSvc')
        self.xmlParserSvc = appMgr.service('XmlParserSvc','IXmlParserSvc')


        p = XYZPoint(0.0,0.0,5000.0)
        v = XYZVector(0.0,0.0,0.0)
        sc = self.magFieldSvc.fieldVector(p, v)
        self.fieldMagnitude = math.sqrt(v.x()*v.x()+v.y()*v.y()+v.z()*v.z())
        self.field = v

        # These are fitting options from $TRACKSYSROOT/TsaSeeeding.opts

        appMgr.property('ToolSvc.TrackmasterFitter').ZPositions = [7500.0, 9450.0, 11900.0]
        appMgr.property('ToolSvc.TrackMasterFitter').StateAtBeamLine = false
        appMgr.property('ToolSvc.TrackMasterFitter').ErrorP          = 0.04
        appMgr.property('ToolSvc.TrackMasterFitter').RefInfoTool = "TsaSeedReferenceCreator";
        appMgr.property('ToolSvc.TrackMasterFitter').SetRefInfo = false
        appMgr.property('ToolSvc.TrackMasterFitter.Extrapolator').ExtraSelector = "TrackSimpleExtraSelector"
        appMgr.property('ToolSvc.TrackMasterFitter.NodeFitter.Extrapolator').ExtraSelector = "TrackSimpleExtraSelector"

        if(self.fieldMagnitude == 0.0):

            # These are fitting options from $TRACKSYSROOT/NoFieldTracking.opts
            print "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"
            print "Running ALIGNT2VELO with Magnetic Field Off TSA refit options"
            print "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"
            
            appMgr.property('ToolSvc.TrackMasterFitter.NodeFitter.Extrapolator').ApplyEnergyLossCorr = false            
            appMgr.property('ToolSvc.TrackMasterFitter').ErrorP = 1.0e-3
            appMgr.property('ToolSvc.TrackMasterFitter.NodeFitter.Extrapolator.ExtraSelector').ExtrapolatorName = "TrackLinearExtrapolator"
            appMgr.property('ToolSvc.TrackMasterFitter.RefInfoTool').SetLRAmbiguities = true



        TrackMasterFitter = appMgr.toolsvc().create('TrackMasterFitter', interface='ITrackFitter')
        self.TrackMasterFitter = TrackMasterFitter


        if(self.numEvProcess < myConstants.minNumberEv):
            print "Requested number of events less than minimum, resetting to minimum of %d"%myConstants.minNumberEv
            self.numEvProcess = myConstants.minNumberEv


        self.OTDetElem = self.det['/dd/Structure/LHCb/AfterMagnetRegion/T/OT']
        self.ITDetElem = self.det['/dd/Structure/LHCb/AfterMagnetRegion/T/IT']
        self.myITCond = self.ITDetElem.geometry().alignmentCondition()
        self.myOTCond = self.OTDetElem.geometry().alignmentCondition()

        self.ITpath = '/Conditions/IT/Alignment/Detectors.xml'
        self.OTpath = '/Conditions/OT/Alignment/Elements.xml'


    def ConnectDB(self,database):


        strbeg = "sqlite_file:"
        strend = "/LHCBCOND"
        dbase = strbeg+database+strend
        self.DTD = '"conddb:/DTD/structure.dtd"'

#        # Only connect to DB if we plan to update it, then a  local copy is required
#        if(myConstants.dbUpdate == 0): return

        print "--- Connecting to Database ---"
        self.database = conddbui.CondDB(dbase, readOnly=False)
        print "---- Done connecting to local DB -----"

    def Analyze(self,numev):

        seedLocation = self.seedlocation
        veloLocation = self.velolocation
        gaudi = self.gaudi

        for i in range(numev):
            print "Processing Event ", i
            gaudi.run(1)
            tseed=self.evt[seedLocation]
            tvelo=self.evt[veloLocation]

            if not tseed:
                print "Non-existing or missing SEED tracks container"
                break
            if not tvelo:
                print "Non-existing or missing VELO tracks container"
                break
                    
            for seed in tseed :

                nIT = 0
                nOT = 0
                meas = seed.lhcbIDs()
            
                for im in meas :
                    if(im.isIT()) :
                        nIT = nIT+1
                    if(im.isOT()) :
                        nOT = nOT+1

                nT = nIT + nOT
                self.nIT.Fill(nIT)
                self.nOT.Fill(nOT)
                self.nT.Fill(nT)

                if(nIT < myConstants.minITHits and nT < myConstants.minOTHits):
                    continue

                sc = self.TrackMasterFitter.fit(seed)

                nm = seed.nLHCbIDs()
                st = seed.closestState(self.zLast)
                tSloX = st.slopes().x()
                tSloY = st.slopes().y()
                tIntX = st.position().x()-tSloX*st.position().z()
                tIntY = st.position().y()-tSloY*st.position().z() 
                ptrk = st.p()/1000.0

                charge = seed.charge() 
                if(tSloX == 0.0):
                    print "Error, Tx Slope = 0: ",nm,tSloX,tSloY,tIntX,tIntY
                    continue
                if(tSloY == 0.0):
                    print "Error, Ty Slope = 0: ",nm,tSloX,tSloY,tIntX,tIntY
                    continue
                
                xp = tIntX+tSloX*8000.0
                yp = tIntY+tSloY*8000.0                
                self.posxy.Fill(xp,yp)
                self.posx.Fill(xp)
                self.posy.Fill(yp)
                
                projSeedX = tSloX*self.zMagnet+tIntX
                projSeedY = tSloY*self.zMagnet+tIntY
            
                phit = atan2(projSeedY, projSeedX)
                           
                iloop = 0
                ivelo = 0
                for velo in tvelo:
                    iloop = iloop + 1
                    ivelo = ivelo + 1
                    nv = velo.nLHCbIDs()
                    sv = velo.closestState(self.zFirst)
                    meas = velo.lhcbIDs()
                    nV = 0  #Count number of R Hits
                    for im in meas :
                        if(im.isVelo()) :
                            nV = nV+1

                    type = velo.type()
                    if(nV < 8): continue
                    vSloX = sv.slopes().x()
                    vSloY = sv.slopes().y()
                    vIntX = sv.position().x()-vSloX*sv.position().z()
                    vIntY = sv.position().y()-vSloY*sv.position().z() 
                    if(vSloX == 0.0):
                        if(iloop==1): print "Error, VeloX Slope = 0: ",ivelo,nV,type,vSloX,vSloY,vIntX,vIntY
                        continue
                    if(vSloY == 0.0):
                        if(iloop==1): print "Error, VeloY Slope = 0: ",ivelo,nv,type,vSloX,vSloY,vIntX,vIntY
                        continue
                    
                    projVeloX = vSloX*self.zMagnet+vIntX
                    projVeloY = vSloY*self.zMagnet+vIntY
                    
                    dxev = projVeloX - projSeedX;
                    dyev = projVeloY - projSeedY;
                    dysev = vSloY - tSloY
                    dxsev = vSloX - tSloX
                    
                    dzxev = -1.0*dxev/tSloX
                    dzyev = -1.0*dyev/tSloY
                
                    phiv = atan2(projVeloY, projVeloX)
                    dphiev = phiv - phit
                    if (dphiev > 2.0*pi) :
                        dphiev = dphiev - 2.0*pi
                    elif (dphiev < -2.0*pi) :
                        dphiev = dphiev + 2.0*pi


                    if (fabs(1000*dysev) < 20.0) : 
                        if(abs(dyev)<50.0):
                            self.dxh.Fill( dxev )
                            if(charge>0): self.dxhp.Fill( dxev )
                            if(charge<0): self.dxhn.Fill( dxev )
                        if(abs(dxev)<50.0):
                            self.dyh.Fill( dyev )
                            if(charge>0): self.dyhp.Fill( dyev )
                            if(charge<0): self.dyhn.Fill( dyev )
                        self.dxsh.Fill( 1000.*dxsev )
                        if(charge>0): self.dxshp.Fill( 1000.*dxsev )
                        if(charge<0): self.dxshn.Fill( 1000.*dxsev )

                        self.dxswh.Fill( 1000.*dxsev )
                        self.dphih.Fill( 1000.*dphiev )
                        if(charge>0): self.dphihp.Fill( 1000.*dphiev )
                        if(charge<0): self.dphihn.Fill( 1000.*dphiev )

                    if(fabs(vSloX) > 0.015) :
                        self.dzxh.Fill( dzxev )
                        self.dzh.Fill( dzxev )
                        if(charge>0): self.dzxhp.Fill( dzxev )
                        if(charge<0): self.dzxhn.Fill( dzxev )
                        if(charge>0): self.dzhp.Fill( dzxev )
                        if(charge<0): self.dzhn.Fill( dzxev )
                    if(fabs(vSloY) > 0.015) :
                        self.dzyh.Fill( dzyev )
                        self.dzh.Fill( dzyev )
                        if(charge>0): self.dzyhp.Fill( dzyev )
                        if(charge<0): self.dzyhn.Fill( dzyev )
                        if(charge>0): self.dzhp.Fill( dzyev )
                        if(charge<0): self.dzhn.Fill( dzyev )
                    if(fabs(dxev)<50.0) :    
                        self.dysh.Fill( 1000.*dysev )
                        if(charge>0): self.dyshp.Fill( 1000.*dysev )
                        if(charge<0): self.dyshn.Fill( 1000.*dysev )
            
                    z = (vIntX-tIntX)/(tSloX-vSloX)
                    if( ptrk > myConstants.pminCut ):
                        self.zinth.Fill( z )
                        if(charge>0): self.zinthp.Fill( z )
                        if(charge<0): self.zinthn.Fill( z )

                    ch = -1.0
                    dangle = tSloX - vSloX
                    if( dangle > 0 ):
                        ch = 1.0
                
                    thxvv = atan(vSloX)
                    pvx = ptrk*sin(thxvv)
                    thxtt = atan(tSloX)
                    ptx = ptrk*sin(thxtt)
                    dpx = ch*(ptx-pvx)
                    if(fabs(dxev)<25 and fabs(dyev)<50 and fabs(dysev)<0.010):
                        self.pkx.Fill( dpx )


        
    def ResetHistograms(self):

        self.dxh.Reset()
        self.dyh.Reset() 
        self.dysh.Reset() 
        self.dxsh.Reset() 
        self.dxswh.Reset() 
        self.dphih.Reset() 
        self.dzxh.Reset() 
        self.dzyh.Reset() 
        self.dzh.Reset() 
        self.zinth.Reset() 
        self.momh.Reset() 
        self.posxy.Reset()
        self.posx.Reset()
        self.posy.Reset() 
        self.nIT.Reset() 
        self.nOT.Reset() 
        self.nT.Reset() 

    def WriteROOTFile(self,root_file):

        if(myConstants.writeAlignFile):

            f=TFile(root_file,'recreate')
            f.Write()
            f.Close()

            #self.dxh.Write()
            #self.dyh.Write() 
            #self.dysh.Write() 
            #self.dxsh.Write() 
            #self.dxswh.Write() 
            #self.dphih.Write() 
            #self.dzxh.Write() 
            #self.dzyh.Write() 
            #self.dzh.Write() 
            #self.zinth.Write() 
            #self.momh.Write() 
            #self.nIT.Write() 
            #self.nOT.Write() 
            #self.nT.Write() 
            
            #f.Close()


    def StoreOffsetsInDB(self,tag):

        myITDetElem = self.ITDetElem
        myOTDetElem = self.OTDetElem
        myDTD       = self.DTD
        db          = self.database
        it_path = self.ITpath
        ot_path = self.OTpath

        myITXml = makeAlignments(myITDetElem , 'IT' , ' ' , myDTD)
        myOTXml = makeAlignments(myOTDetElem , 'OT' , ' ' , myDTD)

        if(myConstants.dbUpdate > 0):
            if(myConstants.dbUpdate == 2 or (myConstants.dbUpdate==1 and myConstants.loadDB) ):
                db.storeXMLString(it_path,{"data":myITXml},conddbui.cool.ValidityKeyMin, conddbui.cool.ValidityKeyMax, 0)
                db.storeXMLString(ot_path,{"data":myOTXml},conddbui.cool.ValidityKeyMin, conddbui.cool.ValidityKeyMax, 0)

        self.updMgrSvc.invalidate(self.myITCond)
        self.updMgrSvc.invalidate(self.myOTCond)
        
        if(myConstants.dbUpdate > 0):
            if(myConstants.dbUpdate == 2 or (myConstants.dbUpdate==1 and myConstants.loadDB) ):
                self.myITCond.forceUpdateMode()
                self.myOTCond.forceUpdateMode()
                db.recursiveTag('/',tag)
                self.condAccSvc.setTag(tag)

                self.xmlParserSvc.clearCache()

        myConstants.loadDB = false
        
        alignCond = myITDetElem.geometry().alignmentCondition()
        thePositions = alignCond.paramAsDoubleVect("dPosXYZ")
        theRotations = alignCond.paramAsDoubleVect("dRotXYZ")
        print "---> The XYZ positions are: ",thePositions[0],thePositions[1],thePositions[2]
        print "---> The XYZ rotations are: ",theRotations[0],theRotations[1],theRotations[2]
        
    def InjectMisalignment(self,t,r,tag):
        myConstants.tglo_trans = t
        myConstants.tglo_rot = r
        self.StoreOffsetsInDB(tag)
    

    def MakeFitHistos3(self):

        opt = 'REBQ0'

        parm = [10.0,   0.001, 0.0,  500.0, 2.0, 500.0, 5.0, 100.0, 15.0]
        range = [-50, 50]
        llim = [0.0,     -5.0, -10000.0, 0.0,  1.0,  0.0, 1.0, 0.0, 1.0]
        ulim = [10000.0,  5.0,  10000.0, 0.0, 20.0,  0.0, 20.0, 0.0, 20.]
        self.dxh = DeltaX
        self.dxh.InputFitPars = parm
        self.dxh.FitRange = range
        self.dxh.LowerParLimit = llim
        self.dxh.UpperParLimit = ulim
        self.dxh.FitOpt = opt
        self.dxh.name = 'Delta X'

        parm = [10.0,   0.001, 0.0,  500.0, 2.0, 500.0,  5.0,  100.0,  15.0]
        range = [-50, 50]
        llim = [0.0,     -5.0, -10000.0, 0.0,  1.0,  0.0,  1.0,  0.0,  1.0]
        ulim = [10000.0,  5.0,  10000.0, 0.0, 20.0,  0.0, 20.0,  0.0, 20.0]
        self.dyh = DeltaY
        self.dyh.InputFitPars = parm
        self.dyh.FitRange = range
        self.dyh.LowerParLimit = llim
        self.dyh.UpperParLimit = ulim
        self.dyh.FitOpt = opt
        self.dyh.name = 'Delta Y'

        parm = [10.0,   0.001, 0.0,  500.0, 0.5, 500.0, 1.5, 100.0, 4.0]
        range = [-20, 20]
        llim = [0.0,     -5.0, -10.0, 0.0,  0.3, 0.0,  0.3, 0.0,  0.3]
        ulim = [10000.0,  5.0,  10.0, 0.0, 10.0, 0.0, 10.0, 0.0, 10.0]
        self.dxsh = DeltaXSlope
        self.dxsh.InputFitPars = parm
        self.dxsh.FitRange = range
        self.dxsh.LowerParLimit = llim
        self.dxsh.UpperParLimit = ulim
        self.dxsh.FitOpt = opt
        self.dxsh.name = 'Delta X Slope'
        
        parm = [10.0,   0.001, 0.0,  100.0, 0.7, 100.0, 2.0, 100.0, 4.0]
        range = [-20, 20]
        llim = [0.0,     -5.0, -10.0, 0.0,  0.5, 0.0, 0.5, 0.0, 0.5]
        ulim = [10000.0,  5.0,  10.0, 0.0, 10.0, 0.0, 10., 0., 10.]
        self.dysh = DeltaYSlope
        self.dysh.InputFitPars = parm
        self.dysh.FitRange = range
        self.dysh.LowerParLimit = llim
        self.dysh.UpperParLimit = ulim
        self.dysh.FitOpt = opt
        self.dysh.name = 'Delta Y Slope'

        
        parm = [10.0,   0.001, 5300,  100.0, 30,  100.0, 30, 50.0, 60]
        range = [5000, 5600]
        llim = [0.0,     -5.0, -10000.0, 0.0,   10.0, 0.0,  10.0, 0.0,  10.0]
        ulim = [10000.0,  5.0,  10000.0, 0.0, 2000.0, 0.0, 2000.0, 0.0, 2000.0]
        self.zinth = Zint
        self.zinth.InputFitPars = parm
        self.zinth.FitRange = range
        self.zinth.LowerParLimit = llim
        self.zinth.UpperParLimit = ulim
        self.zinth.FitOpt = opt
        self.zinth.name = 'Z of Magnet Center'


        
        parm = [10.0, 0.001,  0.0,  5000.0,  8, 5000.0,  20,  1000.0,  50]
        range = [-200, 200]
        llim = [0.0,     -5.0, -10000.0, 0.0,   5.0,  0.0, 5.0,  0.0, 5.0]
        ulim = [10000.0,  5.0,  10000.0, 0.0, 500.0,  0.0, 500.,  0.0, 500.]
        self.dphih = DeltaPhi
        self.dphih.InputFitPars = parm
        self.dphih.FitRange = range
        self.dphih.LowerParLimit = llim
        self.dphih.UpperParLimit = ulim
        self.dphih.FitOpt = opt
        self.dphih.name = 'Delta Phi'
        
        parm = [10.0, 0.001,  0.0001,  100.0,  150.,100.0,  150.,  100.0,     400.]
        range = [-700, 700]
        llim = [0.0,     -5.0, -10000.0, 0.0,  20.0,0.0,  20.0, 0.0,  20.0]
        ulim = [10000.0,  5.0,  10000.0, 0.0, 500.0, 0.0, 500.0, 0.0, 500.0]
        self.dzxh = DeltaZX
        self.dzxh.InputFitPars = parm
        self.dzxh.FitRange = range
        self.dzxh.LowerParLimit = llim
        self.dzxh.UpperParLimit = ulim
        self.dzxh.FitOpt = opt
        self.dzxh.name = 'Delta Z (X view)'


        parm = [10.0, 0.001,  0.0001,  100.0,  150., 100.0,  150.,  100.0, 400.]
        range = [-700, 700]
        llim = [0.0,     -5.0, -10000.0, 0.0,   20.0, 0.0,   20.0, 0.0,   20.0]
        ulim = [10000.0,  5.0,  10000.0, 0.0, 5000.0, 0.0, 5000.0, 0.0, 5000.0]
        self.dzyh = DeltaZY
        self.dzyh.InputFitPars = parm
        self.dzyh.FitRange = range
        self.dzyh.LowerParLimit = llim
        self.dzyh.UpperParLimit = ulim
        self.dzyh.FitOpt = opt
        self.dzyh.name = 'Delta Z (Y view)'

        
        parm = [10.0, 0.001,  0.0001,  100.0,  150., 100.0,  150., 100.0, 400.]
        range = [-700, 700]
        llim = [0.0,     -5.0, -10000.0, 0.0,   20.0,0.0,   20.0,  0.0,   20.0]
        ulim = [10000.0,  5.0,  10000.0, 0.0, 5000.0, 0.0, 5000.0, 0.0, 5000.0]
        self.dzh = DeltaZ
        self.dzh.InputFitPars = parm
        self.dzh.FitRange = range
        self.dzh.LowerParLimit = llim
        self.dzh.UpperParLimit = ulim
        self.dzh.FitOpt = opt
        self.dzh.name = 'Delta Z'


        
    def MakeFitHistos2(self):

        opt = 'REBQ0'

        parm = [10.0,   0.001, 0.0,  500.0,     3.0,  100.0,     8.0]
        range = [-50, 50]
        llim = [0.0,     -5.0, -10000.0, 0.0,  0.5,  0.0, 0.5]
        ulim = [10000.0,  5.0,  10000.0, 0.0, 20.0,  0.0, 20.0]
        self.dxh.InputFitPars = parm
        self.dxh.FitRange = range
        self.dxh.LowerParLimit = llim
        self.dxh.UpperParLimit = ulim
        self.dxh.FitOpt = opt
        self.dxh.name = 'Delta X'

        parm = [10.0,   0.001, 0.0,  500.0,     2.0,  100.0,     10.0]
        range = [-50, 50]
        llim = [0.0,     -5.0, -10000.0, 0.0,    0.5,  0.0,  0.5]
        ulim = [10000.0,  5.0,  10000.0, 0.0, 20.0,    0.0, 20.0]
        self.dyh.InputFitPars = parm
        self.dyh.FitRange = range
        self.dyh.LowerParLimit = llim
        self.dyh.UpperParLimit = ulim
        self.dyh.FitOpt = opt
        self.dyh.name = 'Delta Y'

        parm = [10.0,   0.001, 0.0,  500.0,     0.8,  100.0,     3.0]
        range = [-20, 20]
        llim = [0.0,     -5.0, -10.0, 0.0,  0.3, 0.0,    0.3]
        ulim = [10000.0,  5.0,  10.0, 0.0, 10.0, 0.0,   10.0]
        self.dxsh.InputFitPars = parm
        self.dxsh.FitRange = range
        self.dxsh.LowerParLimit = llim
        self.dxsh.UpperParLimit = ulim
        self.dxsh.FitOpt = opt
        self.dxsh.name = 'Delta X Slope'

        parm = [10.0,   0.001, 0.0,  100.0,     0.8,  100.0,     3.0]
        range = [-20, 20]
        llim = [0.0,     -5.0, -10.0, 0.0,  0.5, 0.0, 0.5]
        ulim = [10000.0,  5.0,  10.0, 0.0, 10.0, 0.0, 10.]
        self.dysh.InputFitPars = parm
        self.dysh.FitRange = range
        self.dysh.LowerParLimit = llim
        self.dysh.UpperParLimit = ulim
        self.dysh.FitOpt = opt
        self.dysh.name = 'Delta Y Slope'

        parm = [10.0,   0.001, 5300,  100.0,     30,  50.0,     60]
        range = [5000, 5600]
        llim = [0.0,     -5.0, -10000.0, 0.0,    1.0, 0.0,   1.0]
        ulim = [10000.0,  5.0,  10000.0, 0.0, 2000.0, 0.0, 2000.0]
        self.zinth.InputFitPars = parm
        self.zinth.FitRange = range
        self.zinth.LowerParLimit = llim
        self.zinth.UpperParLimit = ulim
        self.zinth.FitOpt = opt
        self.zinth.name = 'Z of Magnet Center'

        parm = [10.0, 0.001,  0.0,  5000.0,  20,  1000.0,  50]
        range = [-150, 150]
        llim = [0.0,     -5.0, -10000.0, 0.0,   2.0,  0.0, 2.0]
        ulim = [10000.0,  5.0,  10000.0, 0.0, 500.0,  0.0, 500.]
        self.dphih.InputFitPars = parm
        self.dphih.FitRange = range
        self.dphih.LowerParLimit = llim
        self.dphih.UpperParLimit = ulim
        self.dphih.FitOpt = opt
        self.dphih.name = 'Delta Phi'

        parm = [10.0, 0.001,  0.0001,  100.0,  150.,  100.0,     400.]
        range = [-700, 700]
        llim = [0.0,     -5.0, -10000.0, 0.0,  20.0, 0.0,  20.0]
        ulim = [10000.0,  5.0,  10000.0, 0.0, 500.0, 0.0, 500.0]
        self.dzxh.InputFitPars = parm
        self.dzxh.FitRange = range
        self.dzxh.LowerParLimit = llim
        self.dzxh.UpperParLimit = ulim
        self.dzxh.FitOpt = opt
        self.dzxh.name = 'Delta Z (X view)'

        parm = [10.0, 0.001,  0.0001,  100.0,  150.,  100.0,     400.]
        range = [-700, 700]
        llim = [0.0,     -5.0, -10000.0, 0.0,   20.0, 0.0,   20.0]
        ulim = [10000.0,  5.0,  10000.0, 0.0, 5000.0, 0.0, 5000.0]
        self.dzyh.InputFitPars = parm
        self.dzyh.FitRange = range
        self.dzyh.LowerParLimit = llim
        self.dzyh.UpperParLimit = ulim
        self.dzyh.FitOpt = opt
        self.dzyh.name = 'Delta Z (Y view)'

        parm = [10.0, 0.001,  0.0001,  100.0,  150.,  100.0,     400.]
        range = [-700, 700]
        llim = [0.0,     -5.0, -10000.0, 0.0,   20.0, 0.0,   20.0]
        ulim = [10000.0,  5.0,  10000.0, 0.0, 5000.0, 0.0, 5000.0]
        self.dzh.InputFitPars = parm
        self.dzh.FitRange = range
        self.dzh.LowerParLimit = llim
        self.dzh.UpperParLimit = ulim
        self.dzh.FitOpt = opt
        self.dzh.name = 'Delta Z'

    def Fit(self):

        bfield = job.fieldMagnitude

        hist = self.dxh
        theFitPar = self.fitit(hist)
        value = theFitPar['value']
        error = theFitPar['error']        
        self.dxh.parFit = value
        self.dxh.eparFit = error

        hist = self.dyh
        theFitPar = self.fitit(hist)
        value = theFitPar['value']
        error = theFitPar['error']        
        self.dyh.parFit = value
        self.dyh.eparFit = error


        self.dxsh.parFit = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        self.dxsh.eparFit = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        if(bfield == 0):
            hist = self.dxsh
            theFitPar = self.fitit(hist)
            value = theFitPar['value']
            error = theFitPar['error']        
            self.dxsh.parFit = value
            self.dxsh.eparFit = error


        hist = self.dysh
        theFitPar = self.fitit(hist)
        value = theFitPar['value']
        error = theFitPar['error']        
        self.dysh.parFit = value
        self.dysh.eparFit = error


        hist = self.dphih
        theFitPar = self.fitit(hist)
        value = theFitPar['value']
        error = theFitPar['error']        
        self.dphih.parFit = value
        self.dphih.eparFit = error


        hist = self.dzxh
        theFitPar = self.fitit(hist)
        value = theFitPar['value']
        error = theFitPar['error']        
        self.dzxh.parFit = value
        self.dzxh.eparFit = error

        self.zinth.parFit = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        self.zinth.eparFit = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        self.zinth.parFit[2] = job.zMagnet
        if(bfield != 0):
            hist = self.zinth
            theFitPar = self.fitit(hist)
            value = theFitPar['value']
            error = theFitPar['error']        
            self.zinth.parFit = value
            self.zinth.eparFit = error


    def fitit(self,hist):

        print '-------------------------------------------------'
        print '       Fitting Histogram ',hist.name

        fitFunc = self.fitFunc
        np = fitFunc.GetNumberFreeParameters()
        
        # Update estimated mean with maximum bin position and areas using #entries
        ib = hist.GetMaximumBin()
        m = hist.GetBinCenter(ib)
        hist.InputFitPars[2] = m
        nEntry = hist.GetEntries()
        hist.InputFitPars[3] = 0.3*nEntry
        hist.InputFitPars[5] = 0.3*nEntry
        if (np > 7):
            hist.InputFitPars[7] = 0.2*nEntry
            
        fitFunc.SetRange(hist.FitRange[0],hist.FitRange[1])
        for ip in range(np):
            #print "Par Input/Limits ",hist.InputFitPars[ip]," ",hist.LowerParLimit[ip]," ",hist.UpperParLimit[ip]
            fitFunc.SetParameter(ip,hist.InputFitPars[ip])
            fitFunc.SetParLimits(ip,hist.LowerParLimit[ip],hist.UpperParLimit[ip])

        hist.Fit(fitFunc,hist.FitOpt)

        parFit = fitFunc.GetParameters()
        eparFit = fitFunc.GetParErrors()

        theFitPar = {}
        theFitPar['value']=std.vector('double')()
        theFitPar['error']=std.vector('double')()
        value = theFitPar['value']
        error = theFitPar['error']

        for i in range(np):
            value.push_back(parFit[i])
            error.push_back(eparFit[i])
            

        print "      Mean = %f"%parFit[2],"+- %f"%eparFit[2]
        print '-------------------------------------------------'
#        hist.Draw('e1')

        return theFitPar

    def CheckGlobalPars(self):

        status = 0
        im = 2

        dx = self.dxh.parFit[im]
        dy = self.dyh.parFit[im]
        dz = self.dzxh.parFit[im]
        da = self.dysh.parFit[im]/1000.
        db = self.dxsh.parFit[im]/1000.
        dg = self.dphih.parFit[im]/1000.
        zm = self.zinth.parFit[im]

        edx = self.dxh.eparFit[im]
        edy = self.dyh.eparFit[im]
        edz = self.dzxh.eparFit[im]
        eda = self.dysh.eparFit[im]/1000.
        edb = self.dxsh.eparFit[im]/1000.
        edg = self.dphih.eparFit[im]/1000.
        ezm = self.zinth.eparFit[im]

        myConstants.transPars = [dx, dy, dz]
        myConstants.rotPars = [0.0, 0.0, dg]

        myConstants.tglo_trans = [0.0, 0.0, 0.0]
        myConstants.tglo_rot = [0.0, 0.0, 0.0]
        myConstants.tglo_lasttrans = [0.0, 0.0, 0.0]
        myConstants.tglo_lastrot = [0.0, 0.0, 0.0]

        bfield = self.fieldMagnitude

        if(bfield != 0):
            dev = abs(zm - self.zMagnet)
            devs = abs(zm - self.zMagnet)/ezm
            if(abs(dev) > myConstants.max_dzmag_value and abs(devs)> myConstants.max_dzmags_value):
                print "WARNING: Fitted Z Magnet Center is far from nominal value, fit value = %f %f"%zm,ezm, " - Aborting"
                status = -1
        
        if( abs(dx)> myConstants.max_dx_value ):
            print "X Displacement is unphysically large, ignored, Value = %f"%dx," - Aborting"
            status = -1
        if( abs(dy)> myConstants.max_dy_value ):
            print "Y Displacement is unphysically large, ignored, Value = %f"%dy," - Aborting"
            status = -1
        if( abs(dz)> myConstants.max_dz_value ):
            print "Z Displacement is unphysically large, ignored, Value = %f"%dz," - Aborting"
            status = -1                        
        if( abs(da)> myConstants.max_da_value ):
            print "X Rotation is unphysically large, ignored, Value = %f"%da," - Aborting"
            status = -1            

        if( bfield == 0 ):
            if( abs(db)> myConstants.max_db_value ):
                print "Y Rotation is unphysically large, ignored, Value = %f"%db," - Aborting"
                status = -1
        if( abs(dg)> myConstants.max_dg_value ):
            print "Z Rotation is unphysically large, ignored, Value = %f"%dg," - Aborting"
            status = -1

        if( status < 0 ):    
            return status
            
        # Check if it is within expected error or within tolerance
        # First check that the slopes are within tolerance. If not, fix them, and do a second iteration
        # Note: extra minus sign for X rotations is necessary by coordinate system definition

        myConstants.tglo_etrans[0]=self.dxh.eparFit[im]
        myConstants.tglo_etrans[1]=self.dyh.eparFit[im]
        myConstants.tglo_etrans[2]=self.dzxh.eparFit[im]               
        myConstants.tglo_erot[0]=eda
        myConstants.tglo_erot[2]=edg                

        myConstants.tglo_lasttrans[0]=dx
        myConstants.tglo_lasttrans[1]=dy
        myConstants.tglo_lasttrans[2]=dz
        myConstants.tglo_lastrot[0]=da
        myConstants.tglo_lastrot[1]=db
        myConstants.tglo_lastrot[2]=dg

        if( (abs(da) > myConstants.accept_da_value) and (abs(da/eda) > myConstants.accept_das_value) ):
            print "VELO and T Y Slopes are mismatched, updating CondDB, and proceeding"
            status = 1
            myConstants.fitflags[3] = 0
            myConstants.tglo_rot[0]=-1.0*da
            myConstants.tglo_rot_tot[0] = myConstants.tglo_rot_tot[0] + myConstants.tglo_rot[0]
        elif (myConstants.fitflags[3] == 0):
            myConstants.fitflags[3] = 1
            status = 0
                
        if(bfield == 0):
            myConstants.tglo_erot[1]=edb                
            if( (abs(db) > myConstants.accept_db_value) and (abs(db/edb) > myConstants.accept_dbs_value) ):
                print "VELO and T X Slopes are mismatched, updating CondDB, and proceeding"
                myConstants.fitflags[4] = 0
                status = 1
                myConstants.tglo_rot[1]=1.0*db
                myConstants.tglo_rot_tot[1] = myConstants.tglo_rot_tot[1] + myConstants.tglo_rot[1]
            elif (myConstants.fitflags[4] == 0):
                myConstants.fitflags[4] = 1
                status = 0


        if ( status == 1 ) :
            return status

        if( (abs(dx) > myConstants.accept_dx_value) and (abs(dx/edx) > myConstants.accept_dxs_value) ):
            status = 1
            myConstants.fitflags[0] = 0
            myConstants.tglo_trans[0]=self.dxh.parFit[im]
            myConstants.tglo_trans_tot[0] = myConstants.tglo_trans_tot[0] + myConstants.tglo_trans[0]
        elif (myConstants.fitflags[0] == 0):
            myConstants.fitflags[0] = 1
            status = 0

        if( (abs(dy) > myConstants.accept_dy_value) and (abs(dy/edy) > myConstants.accept_dys_value) ):
            status = 1
            myConstants.fitflags[1] = 0
            myConstants.tglo_trans[1]=self.dyh.parFit[im]
            myConstants.tglo_trans_tot[1] = myConstants.tglo_trans_tot[1] + myConstants.tglo_trans[1]
        elif (myConstants.fitflags[1] == 0):
            myConstants.fitflags[1] = 1
            status = 0

        if( (abs(dz) > myConstants.accept_dz_value) and (abs(dz/edz) > myConstants.accept_dzs_value) ):
            status = 1
            myConstants.fitflags[2] = 0
            myConstants.tglo_trans[2]=self.dzxh.parFit[im]
            myConstants.tglo_trans_tot[2] = myConstants.tglo_trans_tot[2] + myConstants.tglo_trans[2]
        elif (myConstants.fitflags[2] == 0):
            myConstants.fitflags[2] = 1
        
            
        if( (abs(dg) > myConstants.accept_dg_value) and (abs(dg/edg) > myConstants.accept_dgs_value) ):
            status = 1
            myConstants.fitflags[5] = 0
            myConstants.tglo_rot[2]=dg
            myConstants.tglo_rot_tot[2] = myConstants.tglo_rot_tot[2] + myConstants.tglo_rot[2]
        elif (myConstants.fitflags[5] == 0):
            myConstants.fitflags[5] = 1

        if(status == 0):
            myConstants.fitStatus = myConstants.fitStatus + 1

        return status

    def Align(self, filebase, j, tagbase):

        myConstants.fitStatus = 0
        myConstants.fitflags = [ 0, 0, 0, 0, 0, 0]

        
        for i in range(self.numIter):
            
            self.ResetHistograms()
            self.Analyze(numEv)
            self.Fit()
        
            # Write out ROOT file 
            tag = tagbase + "_%1d"%j + "_%1d"%i
            root_file = filebase + "%1d"%j + "_" + str(i) + ".root"
            self.WriteROOTFile(root_file)
            
            # Check fitted parameters
            sc = self.CheckGlobalPars()

            if(self.fieldMagnitude != 0):
                print '----------------------------------------------------------'
                print 'This is a Bfield ON run, Not changing Alignment Constants '
                print 'Writing ROOT file ',root_file,' for monitoring '
                print 'Finishing Job'
                print '----------------------------------------------------------'
                return 1

            #self.PrintInfo()

            if (sc == 0):
                print '----------------------------------------------------------'
                print 'Alignments have converged to an acceptable value, Quitting'
                print '----------------------------------------------------------'
                print 'T Transverse Alignments are: ',myConstants.tglo_trans_tot
                print 'T Rotational Alignments are: ',myConstants.tglo_rot_tot
                myConstants.loadDB = True
            if (sc < 0):
                print 'ERROR with Fitting, Nonsense results '
                print 'Translation Parameters ',myConstants.transPars
                print 'Rotation Parameters ',myConstants.rotPars
                self.sel.rewind()
                return -1
                
            self.StoreOffsetsInDB(tag)

            self.sel.rewind()

            if(sc == 0) :
                return 1

        print '--------------------------------------------------------------'
        print 'Alignments have NOT converged to an acceptable value, Quitting'
        print '--------------------------------------------------------------'
        print 'T Transverse Alignments are: ',myConstants.tglo_trans_tot
        print 'T Rotational Alignments are: ',myConstants.tglo_rot_tot        
        return -1
            
    def SimRandomMisAlign(self,nSim,offset):
        # Generate misalignment values at random
        filebase = 'VeloTAlign_Iter_'        

    
        for j in range(nSim):

            myConstants.tglo_trans = [0.0, 0.0, 0.0]
            myConstants.tglo_rot  = [0.0, 0.0, 0.0]

            tx = myran()*RanCons.sigma_dx
            ty = myran()*RanCons.sigma_dy
            tz = myran()*RanCons.sigma_dz
            rx = myran()*RanCons.sigma_da
            ry = myran()*RanCons.sigma_db
            rz = myran()*RanCons.sigma_dg

            print "----------- Generated Offsets -----------------"
            print "X translation: ",tx,' mm'
            print "Y translation: ",ty,' mm'
            print "Z translation: ",tz,' mm'
            print "X rotation: ",rx,' rad'
            print "Y rotation: ",ry,' rad'
            print "Z rotation: ",rz,' rad'
            print "----------------------------------------------"
    
            t = [tx, ty, tz]
            r = [rx, ry, rz]
            tag = "start1_%1d"%j
            myConstants.loadDB = True
            self.InjectMisalignment(t,r,tag)

            self.tgen = t
            self.rgen = r
        
            myConstants.tglo_trans_tot = [0.0, 0.0, 0.0]
            myConstants.tglo_rot_tot   = [0.0, 0.0, 0.0]
            tagbase = "iter1"
            job.Align(filebase,j,tagbase)
        
            txol = -1.0*myConstants.tglo_lasttrans[0]
            tyol = -1.0*myConstants.tglo_lasttrans[1]
            tzol = -1.0*myConstants.tglo_lasttrans[2]
            rxol = -1.0*myConstants.tglo_lastrot[0]
            ryol = -1.0*myConstants.tglo_lastrot[1]
            rzol = -1.0*myConstants.tglo_lastrot[2]

            txo = -1.0*myConstants.tglo_trans_tot[0]
            tyo = -1.0*myConstants.tglo_trans_tot[1]
            tzo = -1.0*myConstants.tglo_trans_tot[2]
            rxo = -1.0*myConstants.tglo_rot_tot[0]
            ryo = -1.0*myConstants.tglo_rot_tot[1]
            rzo = -1.0*myConstants.tglo_rot_tot[2]
            etxo = 1.0*myConstants.tglo_etrans[0]
            etyo = 1.0*myConstants.tglo_etrans[1]
            etzo = 1.0*myConstants.tglo_etrans[2]
            erxo = 1.0*myConstants.tglo_erot[0]
            eryo = 1.0*myConstants.tglo_erot[1]
            erzo = 1.0*myConstants.tglo_erot[2]
            input = str(tx)+" "+str(ty)+" "+str(tz)+" "+str(rx)+" "+str(ry)+" "+str(rz)+"\n"
            output1 = str(txo)+" "+str(tyo)+" "+str(tzo)+" "+str(rxo)+" "+str(ryo)+" "+str(rzo) + "\n"
            output2 = str(txol)+" "+str(tyol)+" "+str(tzol)+" "+str(rxol)+" "+str(ryol)+" "+str(rzol) + "\n"
            output3 = str(etxo)+" "+str(etyo)+" "+str(etzo)+" "+str(erxo)+" "+str(eryo)+" "+str(erzo) + "\n"

            # Write out to a file
            k = offset + j
            file = "random_misalign_%1d"%k+".dat"
            simExpFile = open(file,'w')
            simExpFile.write(input)
            simExpFile.write(output1)
            simExpFile.write(output2)
            simExpFile.write(output3)
            simExpFile.close()
            
            # shift the DB offsets back to zero
            tx = -1.0*(tx-txo)
            ty = -1.0*(ty-tyo)
            tz = -1.0*(tz-tzo)
            rx = -1.0*(rx-rxo)
            ry = -1.0*(ry-ryo)
            rz = -1.0*(rz-rzo)
            t = [tx, ty, tz]
            r = [rx, ry, rz]
            tag = "end1_%1d"%j
            myConstants.loadDB = True
            self.InjectMisalignment(t,r,tag)



    def SimFixedMisAlign(self,t,r):
        # Generate misalignment values at random
        filebase = 'VeloTAlign_Iter_'        

        self.tgen = t
        self.rgen = r
    
        for j in range(1):

            myConstants.tglo_trans = [0.0, 0.0, 0.0]
            myConstants.tglo_rot  = [0.0, 0.0, 0.0]

            tx = t[0]
            ty = t[1]
            tz = t[2]
            rx = r[0]
            ry = r[1]
            rz = r[2]

            print "----------- Generated Offsets -----------------"
            print "X translation: ",tx,' mm'
            print "Y translation: ",ty,' mm'
            print "Z translation: ",tz,' mm'
            print "X rotation: ",rx,' rad'
            print "Y rotation: ",ry,' rad'
            print "Z rotation: ",rz,' rad'
            print "----------------------------------------------"
    
            tag = "start0_%1d"%j
            myConstants.loadDB = True
            self.InjectMisalignment(t,r,tag)
        
            myConstants.tglo_trans_tot = [0.0, 0.0, 0.0]
            myConstants.tglo_rot_tot   = [0.0, 0.0, 0.0]
            tagbase = "init0"
            job.Align(filebase,j,tagbase)


            txol = -1.0*myConstants.tglo_lasttrans[0]
            tyol = -1.0*myConstants.tglo_lasttrans[1]
            tzol = -1.0*myConstants.tglo_lasttrans[2]
            rxol = -1.0*myConstants.tglo_lastrot[0]
            ryol = -1.0*myConstants.tglo_lastrot[1]
            rzol = -1.0*myConstants.tglo_lastrot[2]

            txo = -1.0*myConstants.tglo_trans_tot[0]
            tyo = -1.0*myConstants.tglo_trans_tot[1]
            tzo = -1.0*myConstants.tglo_trans_tot[2]
            rxo = -1.0*myConstants.tglo_rot_tot[0]
            ryo = -1.0*myConstants.tglo_rot_tot[1]
            rzo = -1.0*myConstants.tglo_rot_tot[2]
            etxo = 1.0*myConstants.tglo_etrans[0]
            etyo = 1.0*myConstants.tglo_etrans[1]
            etzo = 1.0*myConstants.tglo_etrans[2]
            erxo = 1.0*myConstants.tglo_erot[0]
            eryo = 1.0*myConstants.tglo_erot[1]
            erzo = 1.0*myConstants.tglo_erot[2]
            input = str(tx)+" "+str(ty)+" "+str(tz)+" "+str(rx)+" "+str(ry)+" "+str(rz)+"\n"
            output1 = str(txo)+" "+str(tyo)+" "+str(tzo)+" "+str(rxo)+" "+str(ryo)+" "+str(rzo) + "\n"
            output2 = str(txol)+" "+str(tyol)+" "+str(tzol)+" "+str(rxol)+" "+str(ryol)+" "+str(rzol) + "\n"
            output3 = str(etxo)+" "+str(etyo)+" "+str(etzo)+" "+str(erxo)+" "+str(eryo)+" "+str(erzo) + "\n"

            # Write out to a file
            file = "fixed_misalign_%1d"%j+".dat"
            simExpFile = open(file,'w')
            simExpFile.write(input)
            simExpFile.write(output1)
            simExpFile.write(output2)
            simExpFile.write(output3)
            simExpFile.close()
            
            # shift the DB offsets back to zero
            tx = -1.0*(tx-txo)
            ty = -1.0*(ty-tyo)
            tz = -1.0*(tz-tzo)
            rx = -1.0*(rx-rxo)
            ry = -1.0*(ry-ryo)
            rz = -1.0*(rz-rzo)
            t = [tx, ty, tz]
            r = [rx, ry, rz]
            tag = "end0_%1d"%j
            myConstants.loadDB = True
            self.InjectMisalignment(t,r,tag)


    def AlignData(self):
        # Generate misalignment values at random
        filebase = 'VeloTAlignData_Iter_'        
    
        for j in range(1):

            myConstants.tglo_trans = [0.0, 0.0, 0.0]
            myConstants.tglo_rot  = [0.0, 0.0, 0.0]

            myConstants.tglo_trans_tot = [0.0, 0.0, 0.0]
            myConstants.tglo_rot_tot   = [0.0, 0.0, 0.0]
            tagbase = "init0"
            job.Align(filebase,j,tagbase)


            txol = -1.0*myConstants.tglo_lasttrans[0]
            tyol = -1.0*myConstants.tglo_lasttrans[1]
            tzol = -1.0*myConstants.tglo_lasttrans[2]
            rxol = -1.0*myConstants.tglo_lastrot[0]
            ryol = -1.0*myConstants.tglo_lastrot[1]
            rzol = -1.0*myConstants.tglo_lastrot[2]

            txo = -1.0*myConstants.tglo_trans_tot[0]
            tyo = -1.0*myConstants.tglo_trans_tot[1]
            tzo = -1.0*myConstants.tglo_trans_tot[2]
            rxo = -1.0*myConstants.tglo_rot_tot[0]
            ryo = -1.0*myConstants.tglo_rot_tot[1]
            rzo = -1.0*myConstants.tglo_rot_tot[2]
            etxo = 1.0*myConstants.tglo_etrans[0]
            etyo = 1.0*myConstants.tglo_etrans[1]
            etzo = 1.0*myConstants.tglo_etrans[2]
            erxo = 1.0*myConstants.tglo_erot[0]
            eryo = 1.0*myConstants.tglo_erot[1]
            erzo = 1.0*myConstants.tglo_erot[2]
            output1 = str(txo)+" "+str(tyo)+" "+str(tzo)+" "+str(rxo)+" "+str(ryo)+" "+str(rzo) + "\n"
            output2 = str(txol)+" "+str(tyol)+" "+str(tzol)+" "+str(rxol)+" "+str(ryol)+" "+str(rzol) + "\n"
            output3 = str(etxo)+" "+str(etyo)+" "+str(etzo)+" "+str(erxo)+" "+str(eryo)+" "+str(erzo) + "\n"

            # Write out to a file
            file = "data_align_%1d"%j+".dat"
            simExpFile = open(file,'w')
            simExpFile.write(output1)
            simExpFile.write(output2)
            simExpFile.write(output3)
            simExpFile.close()
            

    def PrintInfo(self):
        
        tx = self.tgen[0]
        ty = self.tgen[1]
        tz = self.tgen[2]
        rx = self.rgen[0]
        ry = self.rgen[1]
        rz = self.rgen[2]
        txo = -1.0*myConstants.tglo_trans_tot[0]
        tyo = -1.0*myConstants.tglo_trans_tot[1]
        tzo = -1.0*myConstants.tglo_trans_tot[2]
        rxo = -1.0*myConstants.tglo_rot_tot[0]
        ryo = -1.0*myConstants.tglo_rot_tot[1]
        rzo = -1.0*myConstants.tglo_rot_tot[2]

        print "----------- Generated Offsets -----------------"
        print "X translation: ",tx,' mm'
        print "Y translation: ",ty,' mm'
        print "Z translation: ",tz,' mm'
        print "X rotation: ",rx,' rad'
        print "Y rotation: ",ry,' rad'
        print "Z rotation: ",rz,' rad'
        print "----------------------------------------------"
        print "----------- Reconstructed Offsets -----------------"
        print "X translation: ",txo,' mm'
        print "Y translation: ",tyo,' mm'
        print "Z translation: ",tzo,' mm'
        print "X rotation: ",rxo,' rad'
        print "Y rotation: ",ryo,' rad'
        print "Z rotation: ",rzo,' rad'
        print "----------------------------------------------"

#        print "----------------------------------------------"
#        print "--------- Calculated Offsets ------------------"
#        print "Delta X = ",self.dIntX
#        print "Delta Y = ",self.dIntY
#        print "Delta X Slope = ",self.dSloX
#        print "Delta Y Slope = ",self.dSloY
#        print "----------------------------------------------"        
            



class MyGauss2:

    def __call__(self, x, par):
        g1 = (par[3]/(2.*pi*par[4]))*exp(-0.5*((x[0]-par[2])/par[4])**2)
        g2 = (par[5]/(2.*pi*par[6]))*exp(-0.5*((x[0]-par[2])/par[6])**2)
        return par[0]+par[1]*x[0]+g1+g2

    def __init(self):
        self.SetParameters(10.0,   0.001, 0.0,  500.0,     2.0,  100.0,     10.0)
        self.SetParNames(  "p0","p1", "Mean","Area1","Sigma1","Area2","Sigma2")
        self.SetLineColor(kMagenta)

class MyGauss3:

    def __call__(self, x, par):
        g1 = (par[3]/(2.*pi*par[4]))*exp(-0.5*((x[0]-par[2])/par[4])**2)
        g2 = (par[5]/(2.*pi*par[6]))*exp(-0.5*((x[0]-par[2])/par[6])**2)
        return par[0]+par[1]*x[0]+g1+g2

    def __init(self):
        self.SetParameters(10.0,   0.001, 0.0,  500.0,     2.0,  100.0,     10.0)
        self.SetParNames(  "p0","p1", "Mean","Area1","Sigma1","Area2","Sigma2")
        self.SetLineColor(kMagenta)


def myran():
    r=random.uniform(-1.0,1.0)
    while abs(r)<0.333:
        r=random.uniform(-1.0,1.0)
    return r

def test():
    t = [0.0, 0.0, 0.0]
    r = [0.0, 0.0, 0.0]
    tag = 'i1'
    job.InjectMisalignment(t, r, tag)
    job.Analyze(numEv)

#    job.sel.rewind()
    
#    t = [0.0, 0.0, 0.0]
#    r = [0.002, -0.002, 0.0]
#    tag = 'i2'
#    job.InjectMisalignment(t, r, tag)
#    job.Analyze(1)


if __name__ == '__main__' :

    dbase = "LHCBCOND.db"
#    makeDB(dbase)

    print "Running AlignT2Velo Python Application"
    #--------------------------------------------------------
    #------------------ Setup the Brunel Job ----------------
    #--------------------------------------------------------
    #optsFile = '../options/v200601.opts'
    optsFile = '../options/RunAlignT2VeloFromDST.opts'
    numEv = 1000
    firstEv = 1
    nSim = 1

    procFixed = myConstants.procFixed
    procRandom = myConstants.procRandom
    procData = myConstants.procData
    
    opts,extraopts = getopt.getopt(sys.argv[1:],'f:n:o:s:d:') 
    for o,p in opts:
        if o in ['-f']:
            firstEv = int(p)
        if o in ['-n']:
            numEv = int(p)
        if o in ['-o']:
            optsFile = p
        if o in ['-s']:
            nSim = int(p)
            
    print "------------------------------------------------------------------"
    print " "
    print "          Number of Simulated Experiments = ",nSim
    print "          Number of Events Requested = ",numEv
    print "          First Event Requested = ",firstEv
    print "          Options file = ",optsFile
    print " "
    print "------------------------------------------------------------------"

    job = VeloTAlignment(optsFile)

    gaudi=job.gaudi
    job.configure(seedlocation, velolocation, numEv)
    dbase = 'LHCBCOND.db'
    job.ConnectDB(dbase)

    # simulate a fixed misalignments, t = translations, r=rotations

    test()

    if(procFixed):
        t = myConstants.translation
        r = myConstants.rotation
        job.SimFixedMisAlign(t,r)

    # simulate random misalignments
    if(procRandom):
        offset = 0
        job.SimRandomMisAlign(nSim, offset)

    # Align Data
    if(procData):
        job.AlignData()


        

    
